;; The Fibonacci sequence is defined by the recurrence relation:

;;     Fn = Fn1 + Fn2, where F1 = 1 and F2 = 1.

;; It turns out that F541, which contains 113 digits, is the first Fibonacci number for which the last nine digits are 1-9 pandigital (contain all the digits 1 to 9, but not necessarily in order). And F2749, which contains 575 digits, is the first Fibonacci number for which the first nine digits are 1-9 pandigital.

;; Given that Fk is the first Fibonacci number for which the first nine digits AND the last nine digits are 1-9 pandigital, find k.


(defun fib2 (n f1 f2)
  (cond ((= 0 n)
	 f2)
	(t
	 (fib2 (- n 1) f2 (+ f1 f2)))))

(defun fib (n)
  (cond ((= 0 n) 0)
        ((= 1 n) 1)
        (t
         (fib2 (- n 1) 0 1))))

(defun make-fib-generator ()
  (let ((seq '(1))
	(prev1 0)
	(prev2 1))
    (lambda ()
      (cond ((null seq)
	     (let ((next (+ prev1 prev2)))
	       (setf prev1 prev2)
	       (setf prev2 next)
	       next))
	    (t
	     (prog1 
		 (car seq)
	       (setf seq (cdr seq))))))))

(loop for ii from 1 to 10
      collect (fib ii))

;;(1 1 2 3 5 8 13 21 34 55)

(loop with gen = (make-fib-generator)
      for ii from 1 to 10
      collect (funcall gen))

;;(1 2 3 5 8 13 21 34 55 89)

(defun has-pandigital-start-and-end (num ii)
  (if (< num 100000000)
      (return-from has-pandigital-start-and-end nil))
  ;; we can more quickly test the end by doing a (mod num 1000000000)
  ;; and then testing that,

  ;; there may be something similar we can do to cut down on the
  ;; amount of work that format has to do for the start of the string
  ;; as well...
  (let ((str (format nil "~d" (mod num 1000000000))))
    ;;(format t "str:~a~&" str)
    (if (< (length str) 9)
	(return-from has-pandigital-start-and-end nil))
    (let ((last9  (sort (subseq str (- (length str) 9) (length str)) #'char<)))
      (if (not (equal "123456789" last9))
	  (return-from has-pandigital-start-and-end nil))))
  (format t "found end: ~a~&" ii)
  (let* ((str (format nil "~d" num))
	 (first9 (sort (subseq str 0 9) #'char<)))
    ;;(format t "checking: ~a : ~a : ~a~&"
    ;; str first9 last9)
    (cond ((equal "123456789" first9)
	   (format t "found both: ~a~&" ii)
	   t)
	  (t
	   nil))))


;; (has-pandigital-start-and-end (fib 541) 541)
;; (has-pandigital-start-and-end (fib 2749) 2749)

;; (sort "987654321" #'char<)

;;(has-pandigital-start-and-end 1234567890123498765 'x)

;;(has-pandigital-start-and-end "987123456FEDCBA987654321")

(time 
  (let ((start-at 2749))
    (loop 
     with gen = (make-fib-generator)
     for ii = 1 then (+ ii 1)
     for ff = (funcall gen) then (funcall gen)
     ;; for ii = 2749 then (+ ii 1)
     ;; for ff = (fib ii) then (fib ii)
     ;;repeat 3000
     do
     ;;(format t "ii:~a ff:~a~&" ii ff)
     (when (has-pandigital-start-and-end ff ii)
       (format t "Found: k=~a~&" ii)
       (return ii))
     (when (= 0 (mod ii 1000))
       (format t "ii:~a~&" ii)))))
